#ifndef _SOLVER_H
#define _SOLVER_H

class function
{
public:
	virtual double fun(double *w) = 0 ;
	virtual void grad(double *w, double *g) = 0 ;
	virtual void Hv(double *s, double *Hs) = 0 ;
	virtual void Xv(double *v, double *Xv) = 0;
	virtual void Hessian_coeff(double *C_times_D) = 0;
	virtual double line_search(double *delta_w, double gd, double dTd, double wTd, double *f, double eta, double init_size = 1.0, double *Xd = NULL) = 0 ;

	virtual int get_nr_variable(void) = 0 ;
	virtual ~function(void){}
};

class LBFGS
{
public:
	LBFGS(const function *fun_obj, double eps = 0.1, int m=10, double eta = 1e-2, int max_iter = 100000);
	~LBFGS();

	void lbfgs(double *w);
	void set_print_string(void (*i_print) (const char *buf));

protected:
	void update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** s, double** y, double* g, double* delta, double step_size);
	void TwoLoop(double** inner_product_matrix, double* rho, int DynamicM, int k, int n, double* delta);
	double eps;
	double eta;
	int max_iter;
	int M;
	function *fun_obj;
	void info(const char *fmt,...);
	void (*lbfgs_print_string)(const char *buf);
	int start;
	int length;
	int offset;
};

class NEWTON
{
public:
	NEWTON(const function *fun_obj, double eps = 0.1, double eta = 1e-2, int max_iter = 10000);
	~NEWTON();

	void newton(double *w);
	void set_print_string(void (*i_print) (const char *buf));

private:
	int cg(double *g, double *s, double *r, double cg_eps = 0.1);
	double eps;
	double eta;
	int max_iter;
	function *fun_obj;
	void info(const char *fmt,...);
	void (*newton_print_string)(const char *buf);
};

class TRON
{
public:
	TRON(const function *fun_obj, double eps = 0.1, double default_eps = 0.01, double eps_cg = 0.1, int max_iter = 10000);
	~TRON();

	void tron(double *w);
	void set_print_string(void (*i_print) (const char *buf));

private:
	int trcg(double delta, double *g, double *s, double *r);
	double norm_inf(int n, double *x);

	double eps;
	double eps_cg;
	double default_eps;
	int max_iter;
	function *fun_obj;
	void info(const char *fmt,...);
	void (*tron_print_string)(const char *buf);
};

class MLBFGS: public LBFGS
{
public:
	MLBFGS(const function *fun_obj, int l, double eps = 0.1, int m=5, double eta = 1e-2, int max_iter = 50000);
	~MLBFGS();

	void mlbfgs(double *w);
private:
	void update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** s, double** y, double* g, double* delta, double step_size);
	int l;
	void compute_system(int iter, int k, int n, int DynamicM, double** inner_product_matrix, double** sx, double** yx, double* gx, double* C_times_D, double* system);
};

class MG
{
public:
	MG(const function *fun_obj, int l, double eps = 0.1, int m=10, double eta = 1e-2, int max_iter = 50000);
	~MG();

	void mg(double *w);
	void set_print_string(void (*i_print) (const char *buf));
protected:
	void update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, double** p, double* g);
	int l;
	void compute_system(int iter, int k, int n, int DynamicM, double** inner_product_matrix, double** px, double* C_times_D, double* system);
	double eps;
	double eta;
	int max_iter;
	int M;
	function *fun_obj;
	void info(const char *fmt,...);
	void (*mg_print_string)(const char *buf);
};

class MP : public MG
{
public:
	MP(const function *fun_obj, int l, double eps = 0.1, int m=10, double eta = 1e-2, int max_iter = 50000);

	void mp(double *w);
private:
	void update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** p, double* g, double* delta, double step_size);
};

#endif
